import logging
from datetime import datetime
from typing import Optional
import math

from numpy import ndarray
from pandas import Series, to_datetime
from enum import Enum

from .base import BaseValidator
from ..schema import InferredField, ValidatorTypes

logger = logging.getLogger(__name__)


class TimestampResolution(str, Enum):
    SECONDS = "s"
    MILLISECONDS = "ms"
    NANOSECONDS = "ns"


class Validator(BaseValidator):
    validator_type = ValidatorTypes.NUMERIC
    min_valid_year: int = 1980
    max_valid_year: int = 2100

    @staticmethod
    def _unix_timestamp_resolution(value: float) -> TimestampResolution:
        number_of_digits = len(str(int(value)))
        if number_of_digits == 10:
            return TimestampResolution.SECONDS
        if number_of_digits == 13:
            return TimestampResolution.MILLISECONDS
        if number_of_digits == 16:
            return TimestampResolution.NANOSECONDS

    @staticmethod
    def _fix_input_resolution(value: float, selected_resolution: TimestampResolution) -> float:
        if selected_resolution == TimestampResolution.SECONDS:
            return value
        if selected_resolution == TimestampResolution.MILLISECONDS:
            return value / 1000
        if selected_resolution == TimestampResolution.NANOSECONDS:
            return value / 1000 / 1000

    def validate(self, data: ndarray) -> Optional[InferredField]:
        try:
            self.check_validation_type(dtype=data.dtype)
            selected_resolution = self._unix_timestamp_resolution(float(data[1]))
            if not selected_resolution:
                return None
            for value in data:
                if not math.isnan(value):
                    parsed_value = datetime.utcfromtimestamp(self._fix_input_resolution(value, selected_resolution))
                    if parsed_value.year < self.min_valid_year or parsed_value.year > self.max_valid_year:
                        return None

            return InferredField(
                inferred_type=datetime,
                inferred_pattern=selected_resolution
            )
        except ValueError:
            return None

    def fix(self, column: Series, sample_size: int) -> Series:
        sample_data = column[:sample_size]
        inferred = self.validate(sample_data)
        if inferred:
            return to_datetime(column, unit=inferred.inferred_pattern)



import logging
import os
from importlib import import_module
from types import ModuleType
from typing import List, Optional, Any

from numpy import ndarray

from .schema import Field
from .validators.base import BaseValidator

logger = logging.getLogger(__name__)


class ValidatorsHandler:
    PACKAGE_NAME: str = "shmessy"
    VALIDATORS_DIR: str = "validators"

    def __init__(self):
        self.__validators = self._discover_validators()

    @classmethod
    def _discover_validators(cls) -> List[BaseValidator]:
        validators: List[BaseValidator] = list()
        root_directory = os.path.join(os.path.dirname(__file__))
        validators_directory = os.path.join(root_directory, cls.VALIDATORS_DIR)

        for filename in os.listdir(validators_directory):
            if cls._is_validator(filename):
                validators.append(cls._load_validator(filename).Validator())
        return validators

    @classmethod
    def _load_validator(cls, validator_filename: str) -> Optional[ModuleType]:
        module = f"{cls.PACKAGE_NAME}.{cls.VALIDATORS_DIR}.{validator_filename.replace('.py', '')}"
        try:
            return import_module(module)
        except (ImportError, ValueError, AttributeError) as e:
            logger.error(f"Couldn't import {module}, error is: {e}")

    @staticmethod
    def _is_validator(filename: str) -> bool:
        if "__" in filename or "base" in filename:
            return False
        return True

    def fix_field(self, column: Any, sample_size: int) -> Any:
        for validator in self.__validators:
            fixed_field = validator.fix(column=column, sample_size=sample_size)
            if fixed_field is not None:
                return fixed_field

        return column

    def infer_field(self, field_name: str, data: ndarray) -> Field:

        for validator in self.__validators:
            inferred = validator.validate(data)
            if inferred:
                return Field(
                    field_name=field_name,
                    source_type=data.dtype.type,
                    inferred_type=inferred.inferred_type,
                    inferred_virtual_type=inferred.inferred_virtual_type,
                    inferred_pattern=inferred.inferred_pattern,
                )

        return Field(
            field_name=field_name,
            source_type=data.dtype.type
        )



from enum import Enum
from typing import Optional, List
from typing import Type

from pydantic import BaseModel


class ValidatorTypes(str, Enum):
    NUMERIC = "NUMERIC"
    STRING = "STRING"


class BaseField(BaseModel):
    field_name: str
    source_type: Type


class InferredField(BaseModel):
    inferred_type: Optional[Type] = None
    inferred_virtual_type: Optional[Type] = None
    inferred_pattern: Optional[str] = None


class Field(InferredField, BaseField):
    pass


class ShmessySchema(BaseModel):
    columns: List[Field]



import logging
from datetime import datetime
from typing import Optional

from pandas import DataFrame, to_datetime

from .schema import Field, ShmessySchema, InferredField
from .validators.base import BaseValidator
from .validators_handler import ValidatorsHandler

logger = logging.getLogger(__name__)


class Shmessy:

    def __init__(self, sample_size: Optional[int] = 1000) -> None:
        self.__validators_handler = ValidatorsHandler()
        self.__sample_size = sample_size

    def _get_sampled_df(self, df: DataFrame) -> DataFrame:
        number_of_rows: int = len(df)
        if number_of_rows < self.__sample_size:
            self.__sample_size = number_of_rows
        return df.sample(n=self.__sample_size)

    def infer_schema(self, df: DataFrame) -> ShmessySchema:
        df = self._get_sampled_df(df)

        return ShmessySchema(
            columns=[self.__validators_handler.infer_field(
                field_name=column,
                data=df[column].values
            ) for column in df]
        )

    def fix_schema(self, df: DataFrame) -> DataFrame:
        for column in df:
            df[column] = self.__validators_handler.fix_field(
                column=df[column],
                sample_size=self.__sample_size
            )

        return df



from typing import Optional

from numpy import ndarray
from pandas import Series
from pydantic import EmailStr, BaseModel

from .base import BaseValidator
from ..schema import InferredField, ValidatorTypes


class Model(BaseModel):
    email: EmailStr


class Validator(BaseValidator):
    validator_type = ValidatorTypes.STRING

    def validate(self, data: ndarray) -> Optional[InferredField]:
        try:
            self.check_validation_type(dtype=data.dtype)
        except ValueError:
            return None

        for value in data:
            try:
                Model(email=value)
            except ValueError:
                return None
        return InferredField(
            inferred_type=str,
            inferred_virtual_type=EmailStr
        )

    def fix(self, column: Series, sample_size: int) -> Series:
        sample_data = column[:sample_size]
        inferred = self.validate(sample_data)
        if inferred:
            return column



from abc import ABC, abstractmethod
from typing import Optional, Type, Any
from pandas import Series
from numpy import ndarray, issubdtype, number, object_, str_

from ..schema import InferredField, ValidatorTypes


class BaseValidator(ABC):
    validator_type: ValidatorTypes

    @abstractmethod
    def validate(self, data: ndarray) -> Optional[InferredField]:
        pass

    @abstractmethod
    def fix(self, column: Series, sample_size: int) -> Series:
        pass

    def check_validation_type(self, dtype: Type) -> None:
        if self.validator_type == ValidatorTypes.NUMERIC and not issubdtype(dtype, number):
            raise ValueError(f"NUMERIC validation is not supported for {dtype}")

        if (self.validator_type == ValidatorTypes.STRING and
                not (issubdtype(dtype, object_) or issubdtype(dtype, str_))):
            raise ValueError(f"STRING validation is not supported for {dtype}")



from datetime import datetime
from typing import Optional

from numpy import ndarray
from pandas import Series, to_datetime

from .base import BaseValidator
from ..schema import InferredField, ValidatorTypes


class Validator(BaseValidator):
    validator_type = ValidatorTypes.STRING
    ignore_nan: bool = True
    patterns: list[str] = [
        "%m/%d/%y %H:%M:%S", "%m-%d-%y %H:%M:%S",
        "%m/%d/%Y %H:%M:%S", "%m-%d-%Y %H:%M:%S",
        "%Y/%m/%d %H:%M:%S", "%Y-%m-%d %H:%M:%S",
        "%Y-%m-%d %H:%M:%SZ", "%Y-%m-%dT%H:%M:%SZ",
        "%Y-%m-%dT%H:%M:%S.%f", "%Y-%m-%d %H:%M:%S.%fZ", "%Y-%m-%d %H:%M:%S.%f",
        "%Y-%m-%dT%H:%M:%S.%fZ",
    ]

    def validate(self, data: ndarray) -> Optional[InferredField]:
        try:
            self.check_validation_type(dtype=data.dtype)
        except ValueError:
            return None

        for pattern in self.patterns:
            valid: bool = True
            for value in data:
                try:
                    datetime.strptime(value, pattern)
                except ValueError:
                    valid = False
                    break
                except TypeError:
                    if not self.ignore_nan:
                        valid = False
            if valid:
                return InferredField(
                    inferred_type=datetime,
                    inferred_pattern=pattern
                )

    def fix(self, column: Series, sample_size: int) -> Series:
        sample_data = column[:sample_size]
        inferred = self.validate(sample_data)
        if inferred:
            return to_datetime(column, format=inferred.inferred_pattern)
